import logging
import shutil
import time
import subprocess
from typing import List, Tuple

from .base import BaseFilePlugin, PluginFail
from ..common import *


class SyncPlugin(BaseFilePlugin):
    """Sync files and other data between client and server"""
    MARK = b'sync'
    NAME = 'SyncPlugin'
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()
    DIR_CONFIG_SCHEMA = DictEntry('directory', 'Directory, available for sync', False, entries=(
        StringEntry('name', 'Short name for directory', False, 0, 60, 'Some dir'),
        DirEntry('path', 'Directory to store files and data to show', False, '/tmp/dcnnt/sync/files', True, False),
        TemplateEntry('on_done', 'Template of command executed on sync task completion',
                      True, 0, 4096, None, replacements=(Rep('path', 'Path to saved file', True),)),
    ))
    DIR_CONFIG_DEFAULT = (dict(name='Temporary', path='/tmp/dcnnt/sync/files', on_done=None), )
    FILE_CONFIG_SCHEMA = DictEntry('file', 'File, available for sync', False, entries=(
        StringEntry('name', 'Short name for file', False, 0, 60, 'Some dir'),
        FileEntry('path', 'Path to file', False, '/tmp/dcnnt/sync/dcnnt-example.txt', False, False),
        TemplateEntry('on_merge', 'Template of command executed on merge task to merge local and remote files',
                      True, 0, 4096, default='cat "{local}" "{remote}" > "{output}"',
                      replacements=(Rep('local', 'Path to local version of file', True),
                                    Rep('remote', 'Path to remote version of file', True),
                                    Rep('output', 'Path to save merged file', True),)),
        TemplateEntry('on_done', 'Template of command executed on sync task completion',
                      True, 0, 4096, None, replacements=(Rep('path', 'Path to saved file', True),)),
    ))
    FILE_CONFIG_DEFAULT = (dict(name='Example', path='/tmp/dcnnt/sync/dcnnt-example-file.txt',
                                on_merge='cat "{local}" "{remote}" > "{output}"',
                                on_done='true'),)
    CONFIG_SCHEMA = DictEntry('sync.conf.json', 'Common configuration for sync plugin', False, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', True, 1, 0xFFFFFFF, None),
        DirEntry('working_directory', 'Directory to store temporary files',
                 True, '/tmp/dcnnt/sync_tmp', True, False),
        ListEntry('dir', 'List of directories available for sync', False, 0, 0xFFFF,
                  DIR_CONFIG_DEFAULT, entry=DIR_CONFIG_SCHEMA),
        ListEntry('file', 'List of files available for sync', False, 0, 0xFFFF,
                  FILE_CONFIG_DEFAULT, entry=FILE_CONFIG_SCHEMA),
        DictEntry('contacts', 'Contacts sync settings', False, entries=(
            DirEntry('path', 'Directory to store vcard files', False, '/tmp/dcnnt/sync/contacts', True, False),
            IntEntry('backup_count', 'Count of backup files', False, 0, 4096, 3),
            TemplateEntry('on_done', 'Template of command executed on sync task completion',
                          True, 0, 4096, None, replacements=(Rep('path', 'Path to saved file', True),)),
        )),
        DictEntry('messages', 'SMS sync settings', True, entries=(
            DirEntry('path', 'Directory to store messages files', False, '/tmp/dcnnt/sync/messages', True, False),
            IntEntry('backup_count', 'Count of backup files', False, 0, 4096, 3),
            TemplateEntry('on_done', 'Template of command executed on sync task completion',
                          True, 0, 4096, None, replacements=(Rep('path', 'Path to the last saved file', True),)),
        ))
    ))

    def __init__(self, app, handler, device):
        super().__init__(app, handler, device)

    def handle_targets(self, request: RPCRequest):
        """Return list of sync entries to device"""
        sub = request.params.get('sub')
        if not isinstance(sub, str):
            raise PluginFail('No "sub" param in request')
        entries = self.conf((sub, ))
        self.rpc_send(RPCResponse(request.id, tuple(str(i['path']) for i in entries)))

    @staticmethod
    def get_flat_fs(base: str) -> Dict[str, Tuple[str, int, bool, int]]:
        """Get directory subtree as list"""
        res = dict()
        tree_stop_set = {'', '/', os.path.normpath(base)}
        for root, dirs, files in os.walk(base):
            for d in dirs:
                path = os.path.join(root, d)
                name = os.path.relpath(path, base)
                ts = int(os.stat(path).st_mtime * 1000)
                res[name] = name, ts, True, -2
            for f in files:
                path = os.path.join(root, f)
                name = os.path.relpath(path, base)
                ts = int(os.stat(path).st_mtime * 1000)
                res[name] = name, ts, False, -2
                dir_name = os.path.dirname(path)
                # Here timestamp of directory equals timestamp of newest file
                while dir_name not in tree_stop_set:
                    dir_entry = res.get(dir_name)
                    if dir_entry:
                        _, dir_ts, _ = dir_entry
                        if ts > dir_ts:
                            res[dir_name] = dir_name, ts, True, -2
                    dir_name = os.path.dirname(dir_name)
        return res

    @staticmethod
    def rename_with_mark(base: str, name: str, mark: Union[str, int]) -> Optional[str]:
        """Rename directory sync entry using timestamp"""
        path = os.path.join(base, name)
        if not os.path.exists(path):
            return
        directory = os.path.dirname(path)
        filename = os.path.basename(name)
        s = filename.rsplit('.', maxsplit=1)
        if len(s) == 2:
            name_part, extension = s
        else:
            name_part, extension = filename, ''
        for suffix in ('', '-1', '-2', '-3', '-4', '-5'):
            new_path = os.path.join(directory, f'{name_part}-{mark}{suffix}.{extension}')
            if not os.path.exists(new_path):
                os.rename(path, new_path)
                return new_path[len(base):]
        raise PluginFail(f'No safe name to rename "{path}"')

    @staticmethod
    def ensure_removed(base: str, name: str):
        """Delete FS entry if exists"""
        path = os.path.join(base, name)
        if os.path.exists(path):
            if os.path.isdir(path):
                os.rmdir(path)
            else:
                os.unlink(path)

    def handle_dir_list(self, request: RPCRequest):
        """Initialize directory sync session"""
        args = 'data', 'mode', 'path', 'on_conflict', 'on_delete'
        flat_list_c, mode, path, on_conflict, on_delete = map(request.params.get, args)
        for name, value in zip(args[1:], (mode, path, on_conflict, on_delete)):
            if not isinstance(value, str):
                raise PluginFail(f'No correct "{name}" param in request')
        if path not in tuple(str(i['path']) for i in self.conf(('dir', ))):
            raise PluginFail('Unknown target path')
        do_upload = mode in {'upload', 'sync'}
        do_download = mode in {'download', 'sync'}
        do_delete = on_delete == 'delete'
        self.log(f'Process {len(flat_list_c)} names, mode: "{mode}", target: "{path}", '
                 f'on conflict: "{on_conflict}", on delete: "{on_delete}"')
        # Lists of entries by actions
        to_upload, to_download, to_create_c, to_create_s = list(), list(), list(), list()
        to_rename_c, to_rename_s, to_delete_c, to_delete_s = list(), list(), list(), list()
        # Flat data of FS subtree for server and client
        flat_c: Dict[str, Tuple[str, int, bool, int]] = {i[0]: (i[0], i[1], i[2] == -1, i[2]) for i in flat_list_c}
        flat_s: Dict[str, Tuple[str, int, bool, int]] = self.get_flat_fs(path)
        # flat_list_s = tuple(flat_s.values())
        self.log(f'Created local FS flat data: {len(flat_s)} names')
        # Compare FS subtrees
        names_c, names_s = frozenset(flat_c.keys()), frozenset(flat_s.keys())
        names_client_only = names_c - names_s
        names_server_only = names_s - names_c
        names_both = names_c & names_s
        self.log(f'Names compared: client only: {len(names_client_only)},'
                 f' server only: {len(names_server_only)}, in conflict: {len(names_both)}')
        self.log('Client only:')
        for name in sorted(names_client_only):
            self.log(f'    {name}')
        self.log('Server only:')
        for name in sorted(names_server_only):
            self.log(f'    {name}')
        self.log('In conflict:')
        for name in sorted(names_both):
            self.log(f'    {name}')
        # Process 3 groups of names
        for name in names_client_only:
            if do_upload:
                if flat_c[name][2]:
                    to_create_s.append(name)
                else:
                    to_upload.append(name)
            elif do_delete:  # download only and deletion allowed
                to_delete_c.append(name)
        for name in names_server_only:
            if do_download:
                if flat_s[name][2]:
                    to_create_c.append(name)
                else:
                    to_download.append(name)
            elif do_delete:  # upload only and deletion allowed
                to_delete_s.append(name)
        if on_conflict in {'replace', 'new', 'both'}:  # if conflicts ignored - do nothing
            for name in names_both:
                _, ts_c, is_dir_c, crc_c = flat_c[name]
                _, ts_s, is_dir_s, crc_s = flat_s[name]
                if is_dir_c and is_dir_s:  # both dir already exists, just skip
                    continue
                if mode == 'download':  # from server to client
                    to_c_list = to_create_c if is_dir_s else to_download
                    if on_conflict == 'replace':
                        to_delete_c.append(name)
                        to_c_list.append(name)
                    elif on_conflict == 'new':
                        if ts_s > ts_c:
                            to_delete_c.append(name)
                            to_c_list.append(name)
                    elif on_conflict == 'both':
                        to_rename_c.append(name)
                        to_c_list.append(name)
                elif mode == 'upload':  # from client to server - sort of mirror for previous
                    to_s_list = to_create_s if is_dir_c else to_upload
                    if on_conflict == 'replace':
                        to_delete_s.append(name)
                        to_s_list.append(name)
                    elif on_conflict == 'new':
                        if ts_c > ts_s:
                            to_delete_s.append(name)
                            to_s_list.append(name)
                    elif on_conflict == 'both':
                        to_rename_s.append(name)
                        to_s_list.append(name)
                elif mode == 'sync':  # priority for client here
                    to_c_list = to_create_c if is_dir_s else to_download
                    to_s_list = to_create_s if is_dir_c else to_upload
                    if on_conflict == 'replace':  # replace on client
                        to_delete_c.append(name)
                        to_c_list.append(name)
                    elif on_conflict == 'new':
                        if ts_c > ts_s:
                            to_delete_s.append(name)
                            to_s_list.append(name)
                        else:
                            to_delete_c.append(name)
                            to_c_list.append(name)
                    elif on_conflict == 'both':
                        if is_dir_c != is_dir_s:
                            # Don't know what to do in this case
                            raise PluginFail('Dir-file name conflict')
                        if (not is_dir_c) and (not is_dir_s):
                            srv_ts = flat_s[name][1]
                            new_name_srv = self.rename_with_mark(path, name, f'srv-{srv_ts}')
                            to_upload.append(name)
                            to_download.append(new_name_srv)
        # Print some info to logs
        self.log('To upload from client to server:')
        for name in to_upload:
            self.log(f'    {name}')
        self.log('To download from server to client:')
        for name in to_download:
            self.log(f'    {name}')
        self.log('To delete on client:')
        for name in to_delete_c:
            self.log(f'    {name}')
        self.log('To rename on client:')
        for name in to_rename_c:
            self.log(f'    {name}')
        self.log('Dirs to create on client:')
        for name in to_create_c:
            self.log(f'    {name}')
        self.log('To delete on server:')
        for name in to_delete_s:
            self.log(f'    {name}')
        self.log('To rename on server:')
        for name in to_rename_s:
            self.log(f'    {name}')
        self.log('Dirs to create on server:')
        for name in to_create_s:
            self.log(f'    {name}')
        # Do FS modifications on server
        for name in sorted(to_rename_s):
            new_name = self.rename_with_mark(path, name, flat_s[name][1])
            if path:
                self.log(f'Renamed "{name}" -> "{new_name}"')
        for name in reversed(sorted(to_delete_s)):  # reversed order to ensure files removed before parent dirs
            self.ensure_removed(path, name)
            self.log(f'Removed "{name}"')
        for name in to_create_s:
            os.makedirs(os.path.join(path, name), exist_ok=True)
            self.log(f'Created directory "{name}"')
        # Send response to server
        session_id = f'{time.time()}.{id(request)}'
        self.rpc_send(RPCResponse(request.id, dict(upload=to_upload, download=to_download, create=to_create_c,
                                                   delete=to_delete_c, rename=to_rename_c, session=session_id)))

    def handle_dir_upload(self, request: RPCRequest):
        """Process uploading file on dir sync"""
        print(request.params)
        base = request.params.get('path')
        if base not in tuple(str(i['path']) for i in self.conf(('dir',))):
            raise PluginFail('Unknown target path')
        self.receive_file(request, base)

    def handle_dir_download(self, request: RPCRequest):
        """Process downloading file on dir sync"""
        base = request.params.get('path')
        name = request.params.get('name')
        if base not in tuple(str(i['path']) for i in self.conf(('dir',))):
            raise PluginFail('Unknown target path')
        if not isinstance(name, str):
            raise PluginFail('Incorrect arg "name"')
        path = os.path.join(base, name)
        self.send_file(request, path)

    def common_upload_handler(self, entity_type: str, request: RPCRequest):
        """Process any backup uploading"""
        directory = self.conf((entity_type, 'path'))
        backup_count = self.conf((entity_type, 'backup_count'))
        on_done = self.conf((entity_type, 'on_done'))
        fn = request.params['name']
        total = request.params.get('total')
        index = request.params.get('index')
        is_last_entry = (total is None) and (index is None)
        if isinstance(total, int) and isinstance(index, int):
            if total - index == 1:  # indexes starts from 0 on client
                is_last_entry = True
        if backup_count > 0:
            suffixes = tuple(f'.{i}.bak' for i in reversed(range(backup_count))) + ('', )
            for i in range(len(suffixes) - 1):
                src = os.path.join(directory, fn + suffixes[i + 1])
                dst = os.path.join(directory, fn + suffixes[i])
                if os.path.isfile(src):
                    shutil.copy(src, dst)
        res = self.receive_file(request, directory)
        if on_done and is_last_entry:
            command = on_done.format(path=res)
            self.log(f'Execute: "{command}"')
            subprocess.call(command, shell=True)

    def handle_contacts_upload(self, request: RPCRequest):
        """Process contacts backup uploading"""
        return self.common_upload_handler('contacts', request)

    def handle_messages_upload(self, request: RPCRequest):
        """Process messages backup uploading"""
        return self.common_upload_handler('messages', request)

    def ensure_file_syncable(self, path: str) -> Optional[str]:
        """Check if file in list of sync files and return on_merge option, raise exception otherwise"""
        for i in self.conf('file'):
            if path in i['path']:
                return i.get('on_merge')
        raise PluginFail('No file in sync list')

    def handle_file_info(self, request: RPCRequest):
        """Send some info about sync file to client"""
        path = str(request.params['path'])
        self.ensure_file_syncable(path)
        exists = os.path.isfile(path)
        ts = int(os.path.getmtime(path) * 1000 + .5) if exists else 0
        self.rpc_send(RPCResponse(request.id, {'exists': exists, 'ts': ts}))

    def handle_file_upload(self, request: RPCRequest):
        """Process uploading file on file sync"""
        path = str(request.params.get('path'))
        merge_mode = bool(request.params.get('merge'))
        on_merge = self.ensure_file_syncable(path)
        if merge_mode:
            if on_merge is None:
                raise PluginFail('No merge script set')
            random = str(int(time.time() * 1000)) + str(id(path))
            directory = os.path.join(self.conf('working_directory'), random)
            os.makedirs(directory)
            name = os.path.basename(path)
            server_file_path = os.path.join(directory, f'server-{name}')
            client_file_path = os.path.join(directory, f'client-{name}')
            shutil.copy(path, server_file_path)
            self.receive_file_to_path(request, client_file_path)
            command = on_merge.format(local=server_file_path, remote=client_file_path, output=path)
            self.log(f'Execute: "{command}"')
            subprocess.call(command, shell=True)
        else:
            self.receive_file_to_path(request, path)

    def handle_file_download(self, request: RPCRequest):
        """Process downloading file on file sync"""
        path = request.params.get('path')
        self.ensure_file_syncable(path)
        self.send_file(request, path)

    def process_request(self, request: RPCRequest):
        if request.method == 'get_targets':
            self.handle_targets(request)
        elif request.method == 'dir_list':
            self.handle_dir_list(request)
        elif request.method == 'dir_upload':
            self.handle_dir_upload(request)
        elif request.method == 'dir_download':
            self.handle_dir_download(request)
        elif request.method == 'contacts_upload':
            self.handle_contacts_upload(request)
        elif request.method == 'messages_upload':
            self.handle_messages_upload(request)
        elif request.method == 'file_info':
            self.handle_file_info(request)
        elif request.method == 'file_upload':
            self.handle_file_upload(request)
        elif request.method == 'file_download':
            self.handle_file_download(request)



import logging
import subprocess

from .base import Plugin
from ..common import *


class NotificationsPlugin(Plugin):
    """Receive notifications from phone"""
    MARK = b'nots'
    NAME = 'NotificationsPlugin'
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()
    CONFIG_SCHEMA = DictEntry('rcmd.conf.json', 'Configuration for remote commands file', False, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', True, 1, 0xFFFFFFF, None),
        DirEntry('icon_dir', 'Directory to notification icons', True, '$DCNNT_RUNTIME_DIR', True, False),
        TemplateEntry('cmd', 'Template of notification show command',
                      False, 0, 4096, "notify-send -i '{icon}' '{title}' '{text}'", replacements=(
                          Rep('uin', 'UIN of device which send notification', True),
                          Rep('name', 'Name of device which send notification', True),
                          Rep('package', 'Name of Android package  which create notification', True),
                          Rep('icon', 'Path to saved notification icon', True),
                          Rep('title', 'Title of notification', True),
                          Rep('text', 'Main content of notification', True),
                      ))
    ))

    def __init__(self, app, handler, device):
        super().__init__(app, handler, device)

    def main(self):
        while True:
            request = self.rpc_read()
            self.log(request)
            if request is None:
                self.log('No more requests, stop handler')
                return
            cmd = self.conf('cmd')
            if not cmd:
                return
            if request.method == 'notification':
                icon_data = self.read() if request.params.get('packageIcon', False) else None
                if request.params.get('event') == 'posted':
                    text, package = map(request.params.get, ('text', 'package'))
                    title = request.params.get('title', 'NULL')
                    name, uin = self.device.name, self.device.uin
                    if text is None:
                        text = ''
                    icon_path = os.path.join(self.conf('icon_dir'), f'{package}.{self.device.uin}.icon.png')
                    if bool(icon_data):
                        try:
                            open(icon_path, 'wb').write(icon_data)
                        except Exception as e:
                            self.log(e, logging.WARNING)
                    icon = icon_path if icon_data else ''
                    command = cmd.format(uin=uin, name=name, icon=icon, text=text, title=title, package=package)
                    self.log('Execute: "{}"'.format(command))
                    subprocess.call(command, shell=True)



import fnmatch
import logging
import subprocess
from typing import List

from .base import BaseFilePlugin, HandlerExit, HandlerFail
from ..common import *


class FileTransferPlugin(BaseFilePlugin):
    """Receive file from phone"""
    MARK = b'file'
    NAME = 'FileTransferPlugin'
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()
    DEFAULT_SHARED_DIRS = (dict(path='/tmp/dcnnt/files', name='Shared', glob='*', deep=1024), )
    CONFIG_SCHEMA = DictEntry('file.conf.json', 'Common configuration for file transfer plugin', False, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', True, 1, 0xFFFFFFF, None),
        DirEntry('download_directory', 'Directory to save downloaded files', False,
                 '/tmp/dcnnt/files', True, False),
        TemplateEntry('on_download', 'Template of command executed for every saved file',
                      True, 0, 4096, None, replacements=(Rep('path', 'Path to saved file', True),)),
        ListEntry('shared_dirs', 'Directories shared to client', False, 0, 1073741824, DEFAULT_SHARED_DIRS,
                  entry=DictEntry('shared_dirs[]', 'Description of shared directory', False, entries=(
                      DirEntry('path', 'Path to shared directory', False, '/tmp/dcnnt/files', True, False),
                      StringEntry('name', 'Name using for directory instead of path', True, 0, 60, None),
                      StringEntry('glob', 'UNIX glob to filter visible files in directory', False, 0, 1073741824, '*'),
                      IntEntry('deep', 'Recursion deep for subdirectories', False, 1, 1024, 1)
                  ))),
        ListEntry('shared_dirs_external', 'Lists of directories shared to client', True, 0, 1073741824, (),
                  entry=DictEntry('shared_dirs_external[]', 'Description of shared directory list', False, entries=(
                      FileEntry('path', 'Path to list of shared directories', False,
                                '$DCNNT_CONFIG_DIR/shared.list.txt', True, False),
                      StringEntry('glob', 'UNIX glob to filter visible files in directories',
                                  False, 0, 1073741824, '*'),
                      IntEntry('deep', 'Recursion deep for subdirectories', False, 1, 1024, 1)
                  ))),
    ))
    shared_files_index = list()

    @staticmethod
    def check_file_filter(path: str, glob_str: str) -> bool:
        """Check if file allowed for sharing by filter"""
        return fnmatch.fnmatch(path, glob_str)

    def shared_directory_list(self, directory: str, filter_data, max_deep, current_deep):
        """Create information node for one shared directory"""
        res = list()
        try:
            dir_list = os.listdir(directory)
        except (PermissionError, OSError) as e:
            self.log(f'Could not list content of directory "{directory}" ({e})')
            return res
        for name in dir_list:
            path = os.path.join(directory, name)
            if os.path.isdir(path):
                if current_deep < max_deep and max_deep > 0:
                    dir_list = self.shared_directory_list(path, filter_data, max_deep, current_deep + 1)
                    res.append(dict(name=name, node_type='directory', size=len(dir_list), children=dir_list))
            elif os.path.isfile(path):
                if self.check_file_filter(path, filter_data):
                    self.shared_files_index.append(path)
                    index = len(self.shared_files_index) - 1
                    res.append(dict(name=name, node_type='file', size=os.path.getsize(path), index=index))
        return res

    def process_shared_directory(self, path: str, name: Optional[str], glob: str, deep: int,
                                 res: List[Dict[str, Any]], names: Dict[str, int]):
        """Process one shared directory record"""
        if not os.path.isdir(path):
            self.log(f'Shared directory "{path}" not found', logging.WARN)
            return
        if name is None:
            name = os.path.basename(path)
        if name in names:
            names[name] += 1
            name += f' ({names[name]})'
        else:
            names[name] = 0
        dir_list = self.shared_directory_list(path, glob, deep, 1)
        res.append(dict(name=name, node_type='directory', size=len(dir_list), children=dir_list))

    def shared_files_info(self) -> list:
        """Create tree structure of shared directories"""
        self.shared_files_index.clear()
        res = list()
        names = dict()
        for shared_dir in self.conf('shared_dirs'):
            self.process_shared_directory(shared_dir['path'], shared_dir['name'], shared_dir['glob'],
                                          shared_dir.get('deep', 0), res, names)
        for shared_dirs_import in self.conf('shared_dirs_external'):
            import_path, glob = shared_dirs_import['path'], shared_dirs_import['glob']
            deep = shared_dirs_import.get('deep', 0)
            if not os.path.isfile(import_path):
                self.log(f'List of shared directories "{import_path}" not found', logging.INFO)
                continue
            with open(import_path) as f:
                for path in f.read().splitlines(keepends=False):
                    path = Template(path).safe_substitute(self.app.environment)
                    self.process_shared_directory(path, None, glob, deep, res, names)
        return res

    def handle_upload(self, request: RPCRequest):
        """Receive and save file from client"""
        path = self.receive_file(request, self.conf('download_directory'))
        on_download = self.conf('on_download')
        if isinstance(on_download, str):
            command = on_download.format(path=path)
            self.log('Execute: "{}"'.format(command))
            subprocess.call(command, shell=True)

    def handle_list_shared(self, request: RPCRequest):
        """Create shared files info and return as JSON"""
        try:
            result = self.shared_files_info()
        except Exception as e:
            self.logger.exception('[FileTransferPlugin] {}'.format(e))
            result = INTERNAL_ERROR
        self.rpc_send(RPCResponse(request.id, result))

    def handle_download(self, request):
        """Handle try of device to download file from server"""
        try:
            index, size = request.params['index'], request.params['size']
        except KeyError as e:
            self.log('KeyError {}'.format(e), logging.WARN)
        else:
            self.log('Download request is correct')
            if 0 <= index < len(self.shared_files_index):
                path = self.shared_files_index[index]
                self.send_file(request, path, size)
            else:
                self.rpc_send(RPCResponse(request.id, dict(code=1, message='No such index: {}'.format(index))))

    def process_request(self, request: RPCRequest):
        if request.method == 'list':
            self.handle_list_shared(request)
        elif request.method == 'download':
            self.handle_download(request)
        elif request.method == 'upload':
            self.handle_upload(request)



#!/usr/bin/env python3
import os
import sys
import argparse

from .app import DConnectApp


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--doc', help='Print conf doc and exit', action='store_true')
    parser.add_argument('-c', '--configuration-directory', help='Path to configuration directory',
                        default=os.path.join(os.environ['HOME'], '.config', 'dcnnt'))
    parser.add_argument('--pairing-code', help='Set pre-defined pairing code')
    parser.add_argument('mode', choices=('doc', 'foreground', 'fg', 'pair', 'start', 'stop', 'restart'),
                        nargs='?', default='start',
                        help='Mode to run program: doc - just print config documentation and exit, '
                             'foreground (short: fg) - run program in current tty, '
                             'start/stop/restart - run and stop program as daemon')
    args = parser.parse_args(sys.argv[1:])
    if args.mode == 'doc':
        print(str(DConnectApp.CONFIG_SCHEMA))
        print(DConnectApp.CONFIG_SCHEMA.get_default())
    elif args.mode in {'foreground', 'fg'}:
        app = DConnectApp(args.configuration_directory, True)
        app.init()
        app.run()
        input('Press Enter to stop app...\n')
        app.shutdown()
    else:
        app = DConnectApp(args.configuration_directory, False)
        if args.mode == 'start':
            app.check()
            app.init()
            print(f'Starting in background, pidfile: {app.pidfile}')
            app.daemonize()
            app.run()
        elif args.mode == 'pair':
            app.check()
            app.pair(args.pairing_code)
        elif args.mode == 'stop':
            pid = app.stop()
            if pid:
                print(f'Process {pid} stopped')
        elif args.mode == 'restart':
            pid = app.stop()
            if pid:
                print(f'Process {pid} stopped')
            app.check()
            app.init()
            print(f'Restarting, pidfile: {app.pidfile}')
            app.daemonize()
            app.run()
        else:
            print('Incorrect "mode" argument', file=sys.stderr)
            sys.exit(2)



#!/usr/bin/env python3

"""DConnect application class"""

import sys
import signal
import socket
import logging.handlers
from random import randint
from threading import Thread
from socketserver import ThreadingUDPServer, UDPServer

from .device_manager import DeviceManager, Device
from .server_search import ServerSearchHandler
from .tcp_server import DConnectThreadingTCPServer, DConnectHandler
from .plugins import PLUGINS, PluginInitializer
from .common.jsonconf import *
from .common.daemon import Daemon


class DConnectApp(Daemon):
    """Main application class band all components together"""
    CONFIG_SCHEMA = DictEntry('conf.json', 'Main configuration of dconnect server', False, entries=(
        DictEntry('log', 'Logger configuration', False, entries=(
            FileEntry('path', 'Path to first log file', False, '$HOME/.log/dcnnt.log', True, False),
            IntEntry('size', 'Maximum size of log file', False, 1024, 1073741824, 262144),
            IntEntry('count', 'Count of log files', False, 0, 1024, 3),
        )),
        DictEntry('self', 'Configuration of server device', False, entries=(
            IntEntry('uin', 'Unique identifier of device in network', False, 1, 0xFFFFFFF, randint(0, 1024)),
            StringEntry('name', 'Name of device in network', False, 1, 60, socket.gethostname()),
            StringEntry('description', 'Arbitrary description of device', False, 0, 200, ''),
            StringEntry('password', 'Password to access to device from clients', False, 0, 4096,
                        default=''.join(tuple(chr(randint(ord('a'), ord('z'))) for _ in range(10)))),
        )),
        IntEntry('port', 'Port for UDP and TCP sockets', False, 1, 0xFFFF, 5040),
        FileEntry('pidfile', 'Path to pidfile for daemon mode', True, '', False, False)
    ))

    def __init__(self, directory: str, foreground: bool):
        super().__init__()
        self.dev = None
        self.xdg_runtime_dir = '/tmp'
        self.directory = directory
        self.foreground = foreground
        self.environment = self.init_environment()
        self.conf = self.init_conf(os.path.join(directory, 'conf.json'))
        conf_pidfile = self.conf.get('pidfile')
        self.pidfile = conf_pidfile if conf_pidfile else os.path.join(self.xdg_runtime_dir, 'dcnnt.pid')
        self.log = self.init_logger()
        self.dm = self.plugins = self.udp = self.tcp = self.udp_thread = self.tcp_thread = None

    def pair(self, code: Optional[str] = None):
        """Start app in pairing mode, using pre-defined or random (default) pairing code"""
        if code is None:
            code = str(randint(100000, 999999))
        self.log.setLevel(logging.WARNING)
        self.dm = self.init_dm()
        print('App running in pairing mode')
        print(f'Pair code:\n\n    {code[:3]}-{code[3:]}    \n')
        udp = UDPServer(('0.0.0.0', self.conf['port']), ServerSearchHandler)
        udp.app = self
        udp.pairing_code = code
        udp.paired_uin = None
        signal.signal(signal.SIGINT, lambda a, b: Thread(name='Thread-UDP-Main-Shutdown', target=udp.shutdown).start())
        udp.serve_forever(0.25)
        paired_uin = udp.paired_uin
        print(f'Successful pairing with device {paired_uin}' if paired_uin else 'Pairing failed')
        udp.server_close()
        del udp
        sys.exit(0 if paired_uin else 1)

    def init(self):
        """Create various app internal entities"""
        self.dm = self.init_dm()
        self.plugins = self.init_plugins()
        self.udp = self.init_udp()
        self.tcp = self.init_tcp()
        self.udp_thread = self.tcp_thread = None

    def init_environment(self):
        """Load environment variables, add some local variables and set current directory to config dir"""
        env = {k: v for k, v in os.environ.items()}
        xdg_runtime_dir = env.get('XDG_RUNTIME_DIR', os.path.join('/', 'var', 'run', 'user', str(os.getuid())))
        if not os.path.isdir(xdg_runtime_dir):
            xdg_runtime_dir = '/tmp'
            runtime_dir = os.path.join(xdg_runtime_dir, 'dcnnt')
            os.makedirs(runtime_dir, exist_ok=True)
        else:
            runtime_dir = os.path.join(xdg_runtime_dir, 'dcnnt')
            os.makedirs(runtime_dir, exist_ok=True)
        self.xdg_runtime_dir = xdg_runtime_dir
        env['DCNNT_RUNTIME_DIR'] = runtime_dir
        env['DCNNT_CONFIG_DIR'] = self.directory
        return env

    def init_logger(self):
        """Create console and rotating file logger with parameters specified in configuration"""
        conf = self.conf['log']
        logger = logging.getLogger('dcnnt')
        logger.setLevel(logging.DEBUG)
        if conf['count'] > 0:
            logger.addHandler(logging.handlers.RotatingFileHandler(
                conf['path'], maxBytes=conf['size'], backupCount=conf['count']))
        if self.foreground:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        return logger

    def init_conf(self, path):
        """Load configuration from JSON file"""
        res = ConfigLoader(self.environment, path, self.CONFIG_SCHEMA, True).load()
        if isinstance(res, dict):
            info = res['self']
            dev = Device(info['uin'], info['name'], info['description'], 'server', info['password'])
            self.dev = dev
            return res
        else:
            raise ValueError(f'Main configuration error: {res}')

    def init_dm(self):
        """Init device manager"""
        dm = DeviceManager(self, os.path.join(self.directory, 'devices'))
        dm.load()
        return dm

    def init_plugins(self):
        """Init selected plugins and load settings"""
        plugins = dict()
        plugins_dir = os.path.join(self.directory, 'plugins')
        for plg in PLUGINS:
            initer = PluginInitializer(self.environment, plugins_dir, self.log, plg)
            if initer.init_plugin():
                plugins[plg.MARK] = plg
        return plugins

    def init_udp(self):
        """Init and start UDP server"""
        server = ThreadingUDPServer(('0.0.0.0', self.conf['port']), ServerSearchHandler)
        server.app = self
        return server

    def init_tcp(self):
        """Init and start TCP server"""
        server = DConnectThreadingTCPServer(self, ('0.0.0.0', self.conf['port']), DConnectHandler)
        return server

    def on_sigint(self, *args):
        """SIGINT handler"""
        signal.signal(signal.SIGINT, lambda a, b: None)
        self.shutdown()

    def run(self):
        """Start application"""
        if not self.foreground:
            signal.signal(signal.SIGINT, self.on_sigint)
        self.log.info('START APP')
        self.udp_thread = Thread(None, self.udp.serve_forever, 'UDP-Server-Thread')
        self.log.debug('Starting UDP server...')
        self.udp_thread.start()
        self.log.debug('Starting TCP server...')
        self.tcp_thread = Thread(None, self.tcp.serve_forever, 'TCP-Server-Thread')
        self.tcp_thread.start()
        if not self.foreground:
            self.tcp_thread.join()

    def shutdown(self):
        """Stop all threads and whole application"""
        self.log.info('STOP APP')
        self.log.debug('Shutdown UDP and TCP servers')
        self.udp.shutdown()
        self.tcp.shutdown()
        self.log.debug('Waiting UDP server stop...')
        self.udp_thread.join()
        self.log.debug('Waiting TCP server stop...')
        self.tcp_thread.join()
        self.log.debug('Close TCP socket...')
        self.tcp.server_close()
        self.log.debug('Close UDP socket...')
        self.udp.server_close()
        sys.exit(0)



from .base import PluginInitializer
from .file_transfer import FileTransferPlugin
from .remote_commands import RemoteCommandsPlugin
from .notifications import NotificationsPlugin
from .opener import OpenerPlugin
from .sync import SyncPlugin


PLUGINS = FileTransferPlugin, OpenerPlugin, RemoteCommandsPlugin, NotificationsPlugin, SyncPlugin



"""Simple UDP server to response server-search requests from devices"""

import json
from base64 import b64decode, b64encode
from threading import Thread
from socketserver import BaseRequestHandler

from .common import decrypt, encrypt, derive_key


class ServerSearchHandler(BaseRequestHandler):
    """Handle server search UDP requests from devices"""

    @staticmethod
    def unpack_raw_request(raw):
        """Unpack raw JSON-encoded data from UDP datagram"""
        request = json.loads(raw.decode())
        plugin, action, uin, name, role = map(request.pop, ('plugin', 'action', 'uin', 'name', 'role'))
        pair_data = request.get('pair')
        return plugin, action, uin, name, role, pair_data

    def handle(self):
        app, log, ip, socket = self.server.app, self.server.app.log, self.client_address[0], self.request[1]
        try:
            plugin, action, uin, name, role, pair_data = self.unpack_raw_request(self.request[0])
        except UnicodeDecodeError:
            log.warning('Unicode decoding error in UDP request')
        except json.JSONDecodeError as e:
            log.warning('JSON decoding error in UDP request: {}'.format(e))
        except KeyError as e:
            log.warning('Key not found in JSON (UDP request): {}'.format(e))
        else:
            if plugin == 'search' and action in 'request':
                pairing_code = getattr(self.server, 'pairing_code', None)
                info = app.conf['self']
                app.dm.update_device(uin, ip, name, role)
                if pairing_code and pair_data:
                    password = decrypt(b64decode(pair_data), derive_key(pairing_code + str(info['uin']))).decode()
                    if app.dm.update_device_password(uin, password):
                        Thread(name='Thread-UDP-Server-Shutdown', target=self.server.shutdown).start()
                        self.server.paired_uin = uin
                resp_data = dict(plugin='search', action='response', role='server', uin=info['uin'], name=info['name'])
                if pairing_code:
                    resp_data['pair'] = b64encode(encrypt(str(info['password']).encode(),
                                                          derive_key(pairing_code + str(uin)))).decode()
                response = json.dumps(resp_data)
                res = socket.sendto(response.encode(), self.client_address)



import os
import json
from string import Template
from typing import Optional, Iterable, Union, Any, Dict


class ConfEntryBase:
    """Basic class of simple configuration file value"""
    HOME_DIR = os.environ['HOME']

    def __init__(self, name: str, description: str, optional: bool, default):
        self.name, self.description, self.optional = name, description, optional
        self.default = default() if callable(default) else default

    def pre_process(self, value, environment: Optional[Dict[str, str]] = None) -> Any:
        """Process data from JSON before check"""
        return self.default if value is None and self.optional else value

    def check(self, value, environment: Optional[Dict[str, str]] = None) -> Optional[str]:
        """Check if value is OK"""
        raise NotImplementedError

    def get_default(self):
        """Get default value for this entry"""
        return self.default


class IntEntry(ConfEntryBase):
    """Description of integer config entry"""

    def __init__(self, name: str, description: str, optional: bool,
                 min_value: int, max_value: int, default: Optional[int]):
        super().__init__(name, description, optional, default)
        self.min, self.max = min_value, max_value

    def __str__(self):
        return f'{self.name} - integer{", optional" if self.optional else ""}\n' \
               f'    default: {self.default}\n' \
               f'    min: {self.min}\n' \
               f'    max: {self.max}\n' \
               f'    {self.description}\n'

    def check(self, value, environment: Optional[Dict[str, str]] = None):
        if self.optional and value is None:
            return
        if not isinstance(value, int):
            return f'Type of "{self.name}" is {type(value)}, integer expected'
        if value > self.max:
            return f'Value of "{self.name}" ({value}) is more than max value ({self.max})'
        if value < self.min:
            return f'Value of "{self.name}" ({value}) is less than min value ({self.max})'


class StringEntry(ConfEntryBase):
    """Description of string config entry"""

    def __init__(self, name: str, description: str, optional: bool,
                 min_length: int, max_length: int, default: Optional[str]):
        super().__init__(name, description, optional, default)
        self.min_length, self.max_length = min_length, max_length

    def __str__(self):
        return f'{self.name} - string{", optional" if self.optional else ""}\n' \
               f'    default: {self.default}\n' \
               f'    min length: {self.min_length}\n' \
               f'    max length: {self.max_length}\n' \
               f'    {self.description}\n'

    def check(self, value, environment: Optional[Dict[str, str]] = None):
        if self.optional and value is None:
            return
        if not isinstance(value, str):
            return f'Type of "{self.name}" is {type(value)}, integer expected'
        length = len(value)
        if length > self.max_length:
            return f'Length of "{self.name}" ({length}) is more than max ({self.max_length})'
        if length < self.min_length:
            return f'Length of "{self.name}" ({length}) is less than min ({self.max_length})'


class Rep:
    """Description for replaced part of template"""
    
    def __init__(self, name: str, description: str, optional: bool):
        self.name, self.description, self.optional = name, description, optional

    def __str__(self):
        return f'{self.name}{" (optional)" if self.optional else ""} - {self.description}'
        

class TemplateEntry(StringEntry):
    """Description of sting-template config entry"""
    
    def __init__(self, name: str, description: str, optional: bool, min_length: int,
                 max_length: int, default: Optional[str], replacements: Iterable[Rep]):
        super().__init__(name, description, optional, min_length, max_length, default)
        self.replacements = replacements

    def __str__(self):
        replacement_description = '        \n'.join(tuple(map(str, self.replacements)))
        return f'{self.name} - template string{", optional" if self.optional else ""}\n' \
               f'    default: {self.default}\n' \
               f'    min length: {self.min_length}\n' \
               f'    max length: {self.max_length}\n' \
               f'    replacements:\n' \
               f'        {replacement_description}\n' \
               f'    {self.description}\n'

    def check(self, value, environment: Optional[Dict[str, str]] = None):
        res = super().check(value, environment)
        if res is None:
            if value is None and self.optional:
                return
            test_dict = {i.name: '%TEST%' for i in self.replacements}
            try:
                value.format(**test_dict)
            except KeyError as e:
                return f'Template key failed {str(e)}, required keys: {tuple(test_dict.keys())}'
            for key in filter(lambda x: not x.optional, self.replacements):
                if '{' + key.name not in value:
                    return f'Key "{key.name}" not found in template, required keys: {tuple(test_dict.keys())}'
        else:
            return res


class FileEntry(StringEntry):
    """Description of filesystem path config entry"""

    def __init__(self, name: str, description: str, optional: bool, default: Optional[str],
                 make_dirs: bool, exists: bool):
        super().__init__(name, description, optional, 0, 16384, default)
        self.make_dirs, self.exists = make_dirs, exists

    def __str__(self):

        return f'{self.name} - file path{", optional" if self.optional else ""}\n' \
               f'    default: {self.default}\n' \
               f'    min length: {self.min_length}\n' \
               f'    max length: {self.max_length}\n' \
               f'    must exist: {self.exists}\n' \
               f'    directories created automatically: {self.make_dirs}\n' \
               f'    {self.description}\n'

    def pre_process(self, value, environment: Optional[Dict[str, str]] = None):
        return Template(super().pre_process(value, environment)).safe_substitute(environment)

    def check(self, value, environment: Optional[Dict[str, str]] = None):
        res = super().check(value, environment)
        if res is not None:
            return res
        if not os.path.isfile(value):
            if self.exists:
                return f'File "{value}" not found'
            else:
                dir_path = os.path.dirname(value)
                if dir_path == '':
                    dir_path = '.'
                if not os.path.isdir(dir_path):
                    if self.make_dirs:
                        try:
                            os.makedirs(dir_path, exist_ok=True)
                        except OSError as e:
                            return f'Could not create dir "{dir_path}" ({e})'
                    else:
                        return f'Directory not found "{dir_path}"'


class DirEntry(StringEntry):
    """Description of directory path config entry"""

    def __init__(self, name: str, description: str, optional: bool, default: Optional[str],
                 make_dirs: bool, exists: bool):
        super().__init__(name, description, optional, 0, 16384, default)
        self.make_dirs, self.exists = make_dirs, exists

    def __str__(self):

        return f'{self.name} - directory path{", optional" if self.optional else ""}\n' \
               f'    default: {self.default}\n' \
               f'    min length: {self.min_length}\n' \
               f'    max length: {self.max_length}\n' \
               f'    must exist: {self.exists}\n' \
               f'    directories created automatically: {self.make_dirs}\n' \
               f'    {self.description}\n'

    def pre_process(self, value, environment: Optional[Dict[str, str]] = None):
        return Template(super().pre_process(value, environment)).safe_substitute(environment)

    def check(self, value, environment: Optional[Dict[str, str]] = None):
        res = super().check(value, environment)
        if res is not None:
            return res
        if not os.path.isdir(value):
            if self.exists:
                return f'Directory "{value}" not found'
            else:
                if not os.path.isdir(value):
                    if self.make_dirs:
                        try:
                            os.makedirs(value, exist_ok=True)
                        except OSError as e:
                            return f'Could not create dir "{value}" ({e})'
                    else:
                        return f'Directory not found "{value}"'


class ListEntry(ConfEntryBase):
    """Description of string config entry"""

    def __init__(self, name: str, description: str, optional: bool,
                 min_length: int, max_length: int,  default: tuple, entry: ConfEntryBase):
        super().__init__(name, description, optional, default)
        self.min_length, self.max_length, self.entry = min_length, max_length, entry

    def __str__(self):
        return f'{self.name} - string{", optional" if self.optional else ""}\n' \
               f'    default: {self.default}\n' \
               f'    min length: {self.min_length}\n' \
               f'    max length: {self.max_length}\n' \
               f'    {self.description}\n' \
               f'    entries: \n{str(self.entry)}\n'

    def check(self, value, environment: Optional[Dict[str, str]] = None):
        if self.optional and value is None:
            return
        if isinstance(value, tuple):
            value = list(value)
        if not isinstance(value, list):
            return f'Type of "{self.name}" is {type(value)}, tuple or list expected'
        length = len(value)
        if length > self.max_length:
            return f'Length of "{self.name}" ({length}) is more than max ({self.max_length})'
        if length < self.min_length:
            return f'Length of "{self.name}" ({length}) is less than min ({self.max_length})'
        for i in range(len(value)):
            value[i] = self.entry.pre_process(value[i], environment)
            res = self.entry.check(value[i], environment)
            if res is not None:
                return res


class DictEntry(ConfEntryBase):
    """Description of string config entry"""

    def __init__(self, name: str, description: str, optional: bool, entries: Iterable[ConfEntryBase]):
        super().__init__(name, description, optional, None)
        self.entries = entries

    def __str__(self):
        entries_description = '    ' + '\n    '.join(tuple(i.replace('\n', '\n    ') for i in map(str, self.entries)))
        return f'{self.name} - dictionary{", optional" if self.optional else ""}\n' \
               f'    {self.description}\n' \
               f'    entries: \n\n{entries_description}\n'

    def check(self, value, environment: Optional[Dict[str, str]] = None):
        if self.optional and value is None:
            return
        if not isinstance(value, dict):
            return f'Type of "{self.name}" is {type(value)}, dictionary expected'
        for entry in self.entries:
            name = entry.name
            if name not in value and not entry.optional:
                value[name] = entry.get_default()
                # return f'Entry "{entry.name}" not found in "{self.name}" dictionary'
            value[name] = entry.pre_process(value.get(name), environment)
            res = entry.check(value[name], environment)
            if res is not None:
                return res

    def get_default(self):
        return {i.name: i.get_default() for i in self.entries if not i.optional}


class ConfigLoader:
    """Loading and check facility for JSON configs, can also create default one"""

    def __init__(self, environment: Dict[str, str], path: str, schema: DictEntry, create_defult: bool):
        self.environment, self.path, self.schema, self.create_default = environment, path, schema, create_defult

    def load(self) -> Union[dict, str]:
        """Load and check configuration, create default optionally"""
        if not os.path.isfile(self.path):
            if self.create_default:
                try:
                    os.makedirs(os.path.dirname(self.path), exist_ok=True)
                    with open(self.path, 'w') as f:
                        json.dump(self.schema.get_default(), f, indent=2)
                except BaseException as e:
                    return f'Could not write default configuration to file {self.path} ({e})'
            else:
                return f'File {self.path} not found'
        try:
            conf = json.load(open(self.path))
            if not isinstance(conf, dict):
                return f'JSON config file {self.path} must contain dictionary'
            res = self.schema.check(conf, self.environment)
            if res is not None:
                return f'Error in configuration file {self.path}: {res}'
            return conf
        except IOError:
            return f'Could not open file {self.path}'
        except UnicodeDecodeError:
            return f'None UTF-8 file {self.path}'
        except json.JSONDecodeError:
            return f'JSON error in file {self.path}'
        except BaseException as e:
            return f'Unknown error ({e}) in file {self.path}'



import setuptools
from os import path


here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setuptools.setup(
    name='dcnnt',
    version='0.9.0',
    author='cyanomiko',
    author_email='cyanomiko@protonmail.com',
    description='UI-less tool to connect Android phone with desktop',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cyanomiko/dcnnt-py',
    packages=setuptools.find_packages(),
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: File Sharing',
        'Topic :: System :: Networking',
    ],
    keywords=['phone', 'android', 'sync', 'device'],
    python_requires='>=3.6',
    install_requires=['pycryptodome>=3.9.3'],
    entry_points={
        'console_scripts': [
            'dcnnt=dcnnt.dcnnt:main',
        ],
    },
)



import subprocess

from .base import BaseFilePlugin, PluginFail
from ..common import *


class OpenerPlugin(BaseFilePlugin):
    """Open files, URLs and other data from client"""
    MARK = b'open'
    NAME = 'OpenerPlugin'
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()
    FILE_CONFIG_SCHEMA = DictEntry('file', 'Configuration for file opener', False, entries=(
        DirEntry('download_directory', 'Directory to store files and data to show',
                 False, '/tmp/dcnnt/to_open', True, False),
        TemplateEntry('default_cmd', 'Default command to open file', False, 0, 4096, 'xdg-open "{path}"',
                      replacements=(Rep('path', 'Path to saved file', True),)),
    ))
    LINK_CONFIG_SCHEMA = DictEntry('link', 'Configuration for web link opener', False, entries=(
        TemplateEntry('default_cmd', 'Default command to open web URL', False, 0, 4096, 'xdg-open "{url}"',
                      replacements=(Rep('url', 'Web URL (like http://example.com)', True),)),
    ))
    CONFIG_SCHEMA = DictEntry('open.conf.json', 'Common configuration for opener plugin', False, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', True, 1, 0xFFFFFFF, None),
        FILE_CONFIG_SCHEMA, LINK_CONFIG_SCHEMA
    ))

    def handle_open_file(self, request):
        """Receive and show file from client"""
        path = self.receive_file(request, self.conf(('file', 'download_directory')))
        on_download = self.conf(('file', 'default_cmd'))
        command = on_download.format(path=path)
        self.log('Execute: "{}"'.format(command))
        subprocess.call(command, shell=True)

    def handle_open_link(self, request):
        """Open URL received from client"""
        url = request.params.get('link')
        if not isinstance(url, str):
            raise PluginFail('No "link" param in request')
        self.rpc_send(RPCResponse(request.id, dict(code=0, message='OK')))
        command = self.conf(('link', 'default_cmd')).format(url=url)
        self.log('Execute: "{}"'.format(command))
        subprocess.call(command, shell=True)

    def process_request(self, request: RPCRequest):
        if request.method == 'open_file':
            self.handle_open_file(request)
        elif request.method == 'open_link':
            self.handle_open_link(request)
